/*
 * File parsing and importing.
 *
 * Copyright (C) 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "config.h"
#include "imports.h"
#include "program.h"



/* Parser functions and settings. */

extern void yyparse(void);
extern void yyrestart(FILE *);
extern int yydebug;

/* Specialised parser functions. */

extern void reset(const char *dirname, const char *basename);

/* Parsing products. */

extern struct interface *last_interface;

/* Input directory. */

static char *input_dirname = NULL;



/* Obtain the basename and directory name of generated output. */

static int get_output_location(const char *filename, char **dirname, char **basename)
{
  char *ext;

  /* Obtain the basename of the output prefix. */

  *basename = make_basename(filename);

  if (*basename == NULL)
    return 0;

  /* Truncate the basename to remove the extension. */

  ext = strrchr(*basename, '.');
    
  if (ext != NULL)
    *ext = '\0';

  /* Obtain the directory name from the output prefix. */

  if (conf.output_dir == NULL)
    *dirname = make_dirname(filename);
  else
    *dirname = conf.output_dir;

  if (*dirname == NULL)
  {
    free(*basename);
    return 0;
  }

  return 1;
}

/* Obtain a suitable filename from the parsed import details. */

static struct interface *import_interface_file(const char *quoted_filename)
{
  /* Allocate a string for the quoted filename minus quotes. */

  size_t length = strlen(quoted_filename) - 1;
  char filename[length];

  /* For full paths, allocate enough space for both the dirname and the
     filename. */

  char pathname[strlen(input_dirname) + 1 + length];

  /* Strip quotes from the filename token. */

  strncpy(filename, quoted_filename + 1, length - 1);
  filename[length - 1] = '\0';

  /* Introduce the same dirname as the original input file.
     NOTE: This should probably use a search path. */

  if (is_basename(filename))
  {
    sprintf(pathname, "%s/%s", input_dirname, filename);
    return import_file(pathname, 0);
  }
  else
    return import_file(filename, 0);
}

/* Resolve imports in the interface description. */

static int resolve_imports(struct interface *iface)
{
  struct import *imp;

  for (; iface != NULL; iface = iface->tail)
  {
    /* Traverse the imports list to resolve imports. */

    for (imp = iface->imports; imp != NULL; imp = imp->tail)
    {
      imp->iface = import_interface_file(imp->filename);

      if (imp->iface == NULL)
        return 0;
    }
  }

  return 1;
}

/* Populate an interface reference if the base is found. */

static int populate_base(struct interface_ref *base, struct interface *iface)
{
  if (!strcmp(iface->name, base->name))
  {
    base->iface = iface;
    return 1;
  }
  else
    return 0;
}

/* Find a base interface definition. */

static int find_base_in_imports(struct interface_ref *base, struct import *imp)
{
  struct interface *iface;

  for (; imp != NULL; imp = imp->tail)
  {
    for (iface = imp->iface; iface != NULL; iface = iface->tail)
    {
      if (populate_base(base, iface))
        return 1;

      /* Investigate imports related to an interface. */

      if (find_base_in_imports(base, iface->imports))
        return 1;
    }
  }

  return 0;
}

/* Find a base interface in the current file or in imports. */

static int find_base(struct interface_ref *base, struct interface *iface)
{
  struct interface *preceding;

  /* Search preceding interfaces in the same file. */

  for (preceding = iface->tail; preceding != NULL; preceding = preceding->tail)
  {
    if (populate_base(base, preceding))
      return 1;
  }

  /* Search imported files. */

  if (find_base_in_imports(base, iface->imports))
    return 1;

  fprintf(stderr, "Could not locate base interface %s used by interface %s.\n",
          base->name, iface->name);
  return 0;
}

/* Populate any base interfaces by traversing all interfaces and searching for
   the base interface definitions. */

static int populate_bases(struct interface *iface)
{
  struct interface_ref *base;

  /* Traverse all interfaces. */

  for (; iface != NULL; iface = iface->tail)
  {
    /* Check all bases. */

    for (base = iface->bases; base != NULL; base = base->tail)
    {
      if (!find_base(base, iface))
        return 0;
    }
  }

  return 1;
}



/* Return the last interface from the parsing of the given file, or NULL if
   parsing or importing failed. */

struct interface *import_file(const char *filename, int top_level)
{
  struct interface *result;
  FILE *fp;
  char *dirname, *basename;

  /* Obtain output location details for interfaces within a file. */

  if (!get_output_location(filename, &dirname, &basename))
    return NULL;

  /* Record the input dirname using the original input file. */

  if (input_dirname == NULL)
  {
    input_dirname = make_dirname(filename);
    if (input_dirname == NULL)
      return NULL;
  }

  fp = fopen(filename, "r");

  if (fp == NULL)
  {
    fprintf(stderr, "Could not open file: %s\n", filename);
    return NULL;
  }

  /* Reset any specialised parser state and set the prefix to annotate parsing
     products. */

  reset(dirname, basename);

  yyrestart(fp);
  yyparse();
  fclose(fp);

  /* Obtain the parsing products and process them. */

  result = last_interface;

  if (!resolve_imports(result))
    return NULL;

  if (!populate_bases(result))
    return NULL;

  /* Either report basenames of processed files... */

  if (conf.show_filenames)
    printf("%s\n", basename);

  /* Or write files where directed. */

  else
    if (top_level || conf.generate_all)
      write_files(result);

  return result;
}
