/*
 * Provide a server offering a nested counter.
 *
 * Copyright (C) 2020, 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/env.h>
#include <l4/sys/err.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <pthread.h>
#include <pthread-l4.h>
#include <ipc/server.h>

#include "nested_counter_server.h"



/* Component state and methods. */

struct counter
{
  int counter;
  NestedCounter *next;
  ipc_server_config_type config;
};

static struct counter main_counter;

static long counter_increment(ref_NestedCounter _self, int *result)
{
  struct counter *counter = (struct counter *) (_self.ptr);

  counter->counter += 1;
  *result = counter->counter;
  return L4_EOK;
}

static void init_NestedCounter(NestedCounter *obj, struct counter *counter);

static long counter_next(ref_NestedCounter _self, l4_cap_idx_t *counter)
{
  struct counter *this_counter = (struct counter *) (_self.ptr);
  struct counter *next_counter;
  long err;

  if (this_counter->next == NULL)
  {
    this_counter->next = (NestedCounter *) malloc(sizeof(NestedCounter));
    if (this_counter->next == NULL)
      return -L4_ENOMEM;

    next_counter = (struct counter *) malloc(sizeof(struct counter));
    if (next_counter == NULL)
    {
      free(this_counter->next);
      return -L4_ENOMEM;
    }

    init_NestedCounter(this_counter->next, next_counter);

    /* Create a new server endpoint in the same thread. */

    err = ipc_server_add_config(&this_counter->config, NestedCounter,
                                this_counter->next,
                                pthread_l4_cap(pthread_self()));

    if (err)
      return err;
  }

  *counter = this_counter->config.server;
  return L4_EOK;
}



/* Interfaces for this component. */

iface_NestedCounter server_iface_NestedCounter = {
  .increment=counter_increment,
  .next=counter_next
  };

/* Initialisation. */

static void init_NestedCounter(NestedCounter *obj, struct counter *counter)
{
  counter->counter = 0;
  counter->next = NULL;

  /* Reference to state information and an encapsulation. */

  ref_NestedCounter ref = {.ptr=counter};
  obj->ref = ref;
  obj->iface = &server_iface_NestedCounter;
}



int main(void)
{
  NestedCounter obj;

  init_NestedCounter(&obj, &main_counter);

  /* Wait for messages, dispatching to the handler. */

  ipc_server_loop_for(NestedCounter, &obj, "server");

  return 0;
}
